#!/usr/bin/env python
"""Handy classes to emit RST tables with pleasant column wrapping."""

class Column( object ):
    """Defines a column within a table display.  Can wrap text to fit
    within the column's defined size.
    """
    def __init__( self, name, size ):
        self.name= name
        self.size= size
    def pad( self, value, filler=' ' ):
        return value + (self.size-len(value))*filler
    def wrap( self, value, filler=' ' ):
        value= str(value)
        if len(value) <= self.size:
            return [ self.pad(value,filler), ]
        words= value.split()
        lines= []
        line= ""
        while len(words) != 0:
            nw= words.pop(0)
            if len(line) + len(nw) + 1 > self.size:
                lines.append( line )
                while len(nw) > self.size:
                    lines.append( nw[:self.size] )
                    nw= nw[self.size:]
                line= nw
            else:
                line += nw + " "
        if len(line) > 0:
            lines.append( line )
        return [ self.pad(l,filler) for l in lines ]

class Table( object ):
    def __init__( self, *columns ):
        self.columns= columns
    def _sep( self, filler='-' ):
        row= [ c.pad('',filler=filler) for c in self.columns ]
        return "+".join( [""] + row + [""] )
    def header( self ):
        row= ["Class","Attribute","Segment","Position","source","line"]
        yield self._sep('-')
        yield "|".join( [""] + [ c.pad(c.name) for c in self.columns ] + [""] )
        yield self._sep('=')
    def row( self, *data ):
        colLines = [ col.wrap(data) for col,data in zip(self.columns,data) ]
        sizes= [ len(x) for x in colLines ]
        for i in range(max(sizes)):
            row= []
            for col,data in zip(self.columns,colLines):
                if len(data) != 0:
                    row.append( col.pad( data.pop(0) ) )
                else:
                    row.append( col.pad("") )
            yield "|".join( [""] + row + [""] )
        yield self._sep('-')



#!/usr/bin/env python
"""The X12.map package contains **Visitors** that can traverse the :mod:`X12.parse`
structure to transform it from Python objects to some other form.

Often, a parser is built from some non-Python metadata.  There are a number of
conversion tools in :mod:`tools` that produce an :mod:`X12.parse` structure from other
source data.

The :mod:`X12.parse` structure defines the structure of a message type.  It can also
unmarshall text to create a message instance.

Additionally the parser structure can be used for meta-level processing.
Generally this examines the structure of the message, rather than work with
a specific message instance.  The are several meta-level transformations:

    - Emit Source (in Python, usually)

    - Emit a Data Model (in SQL, as well as the Django ORM)

The :mod:`X12.map.source` package emits Python source in one of two forms.

    -   A single, big constuctor expression for the parser.  This will rarely
        compile.

    -   A flattened definition made up of a number of Loop-level constructors.

The :mod:`X12.map.SQL` package emits SQL DDL for the Message structure.  This isn't
terribly useful, but it does build a minimal database design that
is specific to a message type.

The :mod:`X12.map.dj` package emits Django ORM class definitions for the
various kinds of Segments found within message structure definitions.

XXX - Future Expansion: :mod:`X12.map.sa` package can emit SQLAlchemy definitions
for the various kinds of Segments found within message structure definitions.

Django
=========

..  automodule:: X12.map.dj

Source
===========

..  automodule:: X12.map.source

SQL
==========

..  automodule:: X12.map.SQL

"""



#!/usr/bin/env python
"""Claim File Processing Utilities.

There are a number of convenience functions in this module.

..  autofunction:: genClaims

"""

import logging

def genClaims( aFile, segSep="~", eltSep="*" ):
    """Generator for individual claims from a file.
    This will gracefully read either "Expanded" claims or "Compact" claims,
    and generate "Compact" format messages.
    An expanded claim has :samp:`\\n` after each segment separator.
    A claim always begins with :samp:`ISA` followed by the element separator character,
    usually :samp:`*`.

    :param aFile: the name of the file to open, read and close.
    :param segSep: the expected segment separator, by default :samp:`~`.
    :param eltSep: the expected element separator, by default :samp:`*`.
    """
    log= logging.getLogger( "X12.file.genClaims" )
    log.info( "Reading %s", aFile )
    src= open(aFile,"r")
    claimSoFar= []
    for line in src:
        clean= line.strip()
        if len(clean) == 0:
            continue
        if clean[:4] == "ISA" + eltSep:
            if len(claimSoFar) > 0:
                yield "".join( claimSoFar )
                claimSoFar= []
        claimSoFar.append( clean )
    if len(claimSoFar) > 0:
        yield "".join( claimSoFar )
    src.close()



service_type_codes = {
    "1": "Medical Care",
    "2": "Surgical",
    "3": "Consultation",
    "4": "Diagnostic X-Ray",
    "5": "Diagnostic Lab",
    "6": "Radiation Therapy",
    "7": "Anesthesia",
    "8": "Surgical Assistance",
    "9": "Other Medical",
    "10": "Blood",
    "11": "Durable Medical Equipment Used",
    "12": "Durable Medical Equipment Purchased",
    "13": "Ambulatory Service Center Facility",
    "14": "Renal Supplies",
    "15": "Alternate Method Dialysis",
    "16": "Chronic Renal Disease (CRD) Equipment",
    "17": "Pre-Admission Testing",
    "18": "Durable Medical Equipment Rental",
    "19": "Pneumonia Vaccine",
    "20": "Second Surgical Opinion",
    "21": "Third Surgical Opinion",
    "22": "Social Work",
    "23": "Diagnostic Dental",
    "24": "Periodontics",
    "25": "Restorative",
    "26": "Endodontics",
    "27": "Maxillofacial Prosthetics",
    "28": "Adjunctive Dental Services",
    "30": "Health Benefit Plan Coverage",
    "32": "Plan Waiting Period",
    "33": "Chiropractic",
    "34": "Chiropractic Modality",
    "35": "Dental Care",
    "36": "Dental Crowns",
    "37": "Dental Accident",
    "38": "Orthodontics",
    "39": "Prosthodontics",
    "40": "Oral Surgery",
    "41": "Preventive Dental",
    "42": "Home Health Care",
    "43": "Home Health Prescriptions",
    "44": "Home Health Visits",
    "45": "Hospice",
    "46": "Respite Care",
    "47": "Hospitalization",
    "48": "Hospital - Inpatient",
    "49": "Hospital - Room and Board",
    "50": "Hospital - Outpatient",
    "51": "Hospital - Emergency Accident",
    "52": "Hospital - Emergency Medical",
    "53": "Hospital - Ambulatory Surgical",
    "54": "Long Term Care",
    "55": "Major Medical",
    "56": "Medically Related Transportation",
    "57": "Air Transportation",
    "58": "Cabulance",
    "59": "Licensed Ambulance",
    "60": "General Benefits",
    "61": "In-vitro Fertilization",
    "62": "MRI Scan",
    "63": "Donor Procedures",
    "64": "Acupuncture",
    "65": "Newborn Care",
    "66": "Pathology",
    "67": "Smoking Cessation",
    "68": "Well Baby Care",
    "69": "Maternity",
    "70": "Transplants",
    "71": "Audiology",
    "72": "Inhalation Therapy",
    "73": "Diagnostic Medical",
    "74": "Private Duty Nursing",
    "75": "Prosthetic Device",
    "76": "Dialysis",
    "77": "Otology",
    "78": "Chemotherapy",
    "79": "Allergy Testing",
    "80": "Immunizations",
    "81": "Routine Physical",
    "82": "Family Planning",
    "83": "Infertility",
    "84": "Abortion",
    "85": "HIV - AIDS Treatment",
    "86": "Emergency Services",
    "87": "Cancer Treatment",
    "88": "Pharmacy",
    "89": "Free Standing Prescription Drug",
    "90": "Mail Order Prescription Drug",
    "91": "Brand Name Prescription Drug",
    "92": "Generic Prescription Drug",
    "93": "Podiatry",
    "94": "Podiatry - Office Visits",
    "95": "Podiatry - Nursing Home Visits",
    "96": "Professional (Physician)",
    "97": "Anesthesiologist",
    "98": "Professional (Physician) Visit - Office",
    "99": "Professional (Physician) Visit - Inpatient",
    "A0": "Professional (Physician) Visit - Outpatient",
    "A1": "Professional (Physician) Visit - Nursing Home",
    "A2": "Professional (Physician) Visit - Skilled Nursing Facility",
    "A3": "Professional (Physician) Visit - Home",
    "A4": "Psychiatric",
    "A5": "Psychiatric - Room and Board",
    "A6": "Psychotherapy",
    "A7": "Psychiatric - Inpatient",
    "A8": "Psychiatric - Outpatient",
    "A9": "Rehabilitation",
    "AA": "Rehabilitation - Room and Board",
    "AB": "Rehabilitation - Inpatient",
    "AC": "Rehabilitation - Outpatient",
    "AD": "Occupational Therapy",
    "AE": "Physical Medicine",
    "AF": "Speech Therapy",
    "AG": "Skilled Nursing Care",
    "AH": "Skilled Nursing Care - Room and Board",
    "AI": "Substance Abuse",
    "AJ": "Alcoholism Treatment",
    "AK": "Drug Addiction",
    "AL": "Optometry",
    "AM": "Frames",
    "AO": "Lenses",
    "AP": "Routine Eye Exam",
    "AQ": "Nonmedically Necessary Physical (These physicals are required by "
            "other entities e.g., insurance application, pilot license, "
            "employment or school)",
    "AR": "Experimental Drug Therapy",
    "B1": "Burn Care",
    "B2": "Brand Name Prescription Drug - Formulary",
    "B3": "Brand Name Prescription Drug - Non-Formulary",
    "BA": "Independent Medical Evaluation",
    "BB": "Psychiatric Treatment Partial Hospitalization",
    "BC": "Day Care (Psychiatric)",
    "BD": "Cognitive Therapy",
    "BE": "Massage Therapy",
    "BF": "Pulmonary Rehabilitation",
    "BG": "Cardiac Rehabilitation",
    "BH": "Pediatric",
    "BI": "Nursery Room and Board",
    "BJ": "Skin",
    "BK": "Orthopedic",
    "BL": "Cardiac",
    "BM": "Lymphatic",
    "BN": "Gastrointestinal",
    "BP": "Endocrine",
    "BQ": "Neurology",
    "BR": "Eye",
    "BS": "Invasive Procedures",
    "BT": "Gynecological",
    "BU": "Obstetrical",
    "BV": "Obstetrical/Gynecological",
    "BW": "Mail Order Prescription Drug: Brand Name",
    "BX": "Mail Order Prescription Drug: Generic",
    "BY": "Physician Visit - Sick",
    "BZ": "Physician Visit - Well",
    "C1": "Coronary Care",
    "CK": "Screening X-ray",
    "CL": "Screening laboratory",
    "CM": "Mammogram, High Risk Patient",
    "CN": "Mammogram, Low Risk Patient",
    "CO": "Flu Vaccination",
    "CP": "Eyewear Accessories",
    "CQ": "Case Management",
    "DG": "Dermatology",
    "DM": "Durable Medical Equipment",
    "DS": "Diabetic Supplies",
    "E0": "Allied Behavioral Analysis Therapy",
    "E1": "Non-Medical Equipment (non DME)",
    "E2": "Psychiatric Emergency",
    "E3": "Step Down Unit",
    "E4": "Skilled Nursing Facility Head Level of Care",
    "E5": "Skilled Nursing Facility Ventilator Level of Care",
    "E6": "Level of Care 1",
    "E7": "Level of Care 2",
    "E8": "Level of Care 3",
    "E9": "Level of Care 4",
    "E10": "Radiographs",
    "E11": "Diagnostic Imaging",
    "E12": "Basic Restorative - Dental",
    "E13": "Major Restorative - Dental",
    "E14": "Fixed Prosthodontics",
    "E15": "Removable Prosthodontics",
    "E16": "Intraoral Images - Complete Series",
    "E17": "Oral Evaluation",
    "E18": "Dental Prophylaxis",
    "E19": "Panoramic Images",
    "E20": "Sealants",
    "E21": "Flouride Treatments",
    "E22": "Dental Implants",
    "E23": "Temporomandibular Joint Dysfunction",
    "E24": "Retail Pharmacy Prescription Drug",
    "E25": "Long Term Care Pharmacy",
    "E26": "Comprehensive Medication Therapy Management Review",
    "E27": "Targeted Medication Therapy Management Review",
    "E28": "Dietary/Nutritional Services",
    "EA": "Preventive Services",
    "EB": "Specialty Pharmacy",
    "EC": "Durable Medical Equipment New",
    "ED": "CAT Scan",
    "EE": "Ophthalmology",
    "EF": "Contact Lenses",
    "GF": "Generic Prescription Drug - Formulary",
    "GN": "Generic Prescription Drug - Non-Formulary",
    "GY": "Allergy",
    "IC": "Intensive Care",
    "MH": "Mental Health",
    "NI": "Neonatal Intensive Care",
    "ON": "Oncology",
    "PE": "Positron Emission Tomography (PET) Scan",
    "PT": "Physical Therapy",
    "PU": "Pulmonary",
    "RN": "Renal",
    "RT": "Residential Psychiatric Treatment",
    "SMH": "Serious Mental Health",
    "TC": "Transitional Care",
    "TN": "Transitional Nursery Care",
    "UC": "Urgent Care",
}

coverage_level = {
    "CHD": "Children Only",
    "DEP": "Dependents Only",
    "ECH": "Employee and Children",
    "EMP": "Employee Only",
    "ESP": "Employee and Spouse",
    "FAM": "Family",
    "IND": "Individual",
    "SPC": "Spouse and Children",
    "SPO": "Spouse Only"
}

insurance_type = {
    "12": "Medicare Secondary Working Aged Beneficiary or Spouse with "
            "Employer Group Health Plan",
    "13": "Medicare Secondary End-Stage Renal Disease Beneficiary in the 12 "
            "month coordination period with an employer's group health plan",
    "14": "Medicare Secondary, No-fault Insurance including Auto is Primary",
    "15": "Medicare Secondary Worker's Compensation",
    "16": "Medicare Secondary Public Health Service (PHS)or Other Federal "
            "Agency",
    "41": "Medicare Secondary Black Lung",
    "42": "Medicare Secondary Veteran's Administration",
    "43": "Medicare Secondary Disabled Beneficiary Under Age 65 with Large "
            "Group Health Plan (LGHP)",
    "47": "Medicare Secondary, Other Liability Insurance is Primary",
    "AP": "Auto Insurance Policy",
    "C1": "Commercial",
    "CO": "Consolidated Omnibus Budget Reconciliation Act (COBRA)",
    "CP": "Medicare Conditionally Primary",
    "D": "Disability",
    "DB": "Disability Benefits",
    "EP": "Exclusive Provider Organization",
    "FF": "Family or Friends",
    "GP": "Group Policy",
    "HM": "Health Maintenance Organization (HMO)",
    "HN": "Health Maintenance Organization (HMO) - Medicare Risk",
    "HS": "Special Low Income Medicare Beneficiary",
    "IN": "Indemnity",
    "IP": "Individual Policy",
    "LC": "Long Term Care",
    "LD": "Long Term Policy",
    "LI": "Life Insurance",
    "LT": "Litigation",
    "MA": "Medicare Part A",
    "MB": "Medicare Part B",
    "MC": "Medicaid",
    "MH": "Medigap Part A",
    "MI": "Medigap Part B",
    "MP": "Medicare Primary",
    "OT": "Other",
    "PE": "Property Insurance - Personal",
    "PL": "Personal",
    "PP": "Personal Payment (Cash - No Insurance)",
    "PR": "Preferred Provider Organization (PPO)",
    "PS": "Point of Service (POS)",
    "QM": "Qualified Medicare Beneficiary",
    "RP": "Property Insurance - Real",
    "SP": "Supplemental Policy",
    "TF": "Tax Equity Fiscal Responsibility Act (TEFRA)",
    "WC": "Workers Compensation",
    "WU": "Wrap Up Policy",
}

reject_reason_code = {
    "01": "Price Authorization Invalid",
    "02": "Price Authorization Expired",
    "03": "Product not on the price authorization",
    "04": "Authorized Quantity Exceeded",
    "05": "Zero Balance",
    "06": "Special Cost Incorrect",
    "07": "Catalog Cost Incorrect",
    "08": "Invalid Ship Location",
    "09": "No Credit Allowed",
    "10": "Administrative Cancellation",
    "11": "Invalid Debit Number",
    "12": "Duplicate Sequence Number",
    "13": "Not Valid for Price Protection",
    "14": "Invalid part number",
    "15": "Required application data missing",
    "16": "Unit resale higher than authorized",
    "17": "Negotiated price was not less than book price",
    "18": "Ship date must not be after current date",
    "19": "Ship date cannot be prior to price authorization issue date",
    "20": "Ship date should not be before price authorization date (for "
            "rebills)",
    "21": "Price authorization is a rebill type",
    "23": "Price authorization has been deleted",
    "24": "Price authorization used on a sales order",
    "25": "Disposition pending vendor review.",
    "26": "Invalid Customer Number",
    "27": "Invalid Ship Date",
    "28": "Duplicate Invoice Number",
    "29": "Claim Submitted Past Exercise Period",
    "30": "Invalid Meet Competition Cost",
    "31": "Invalid Book Cost",
    "32": "Input Incomplete",
    "33": "Input Errors",
    "34": "No Coverage",
    "35": "Out of Network",
    "36": "Testing not Included",
    "37": "Request Forwarded To and Decision Response Forthcoming From an "
            "External Review Organization",
    "41": "Authorization/Access Restrictions",
    "42": "Unable to Respond at Current Time",
    "43": "Invalid/Missing Provider Identification",
    "44": "Invalid/Missing Provider Name",
    "45": "Invalid/Missing Provider Specialty",
    "46": "Invalid/Missing Provider Phone Number",
    "47": "Invalid/Missing Provider State",
    "48": "Invalid/Missing Referring Provider Identification Number",
    "49": "Provider is Not Primary Care Physician",
    "50": "Provider Ineligible for Inquiries",
    "51": "Provider Not on File",
    "52": "Service Dates Not Within Provider Plan Enrollment",
    "53": "Inquired Benefit Inconsistent with Provider Type",
    "54": "Inappropriate Product/Service ID Qualifier",
    "55": "Inappropriate Product/Service ID",
    "56": "Inappropriate Date",
    "57": "Invalid/Missing Date(s) of Service",
    "58": "Invalid/Missing Date-of-Birth",
    "59": "Invalid/Missing Date-of-Death",
    "60": "Date of Birth Follows Date(s) of Service",
    "61": "Date of Death Precedes Date(s) of Service",
    "62": "Date of Service Not Within Allowable Inquiry Period",
    "63": "Date of Service in Future",
    "64": "Invalid/Missing Patient ID",
    "65": "Invalid/Missing Patient Name",
    "66": "Invalid/Missing Patient Gender Code",
    "67": "Patient Not Found",
    "68": "Duplicate Patient ID Number",
    "69": "Inconsistent with Patient's Age",
    "70": "Inconsistent with Patient's Gender",
    "71": "Patient Birth Date Does Not Match That for the Patient on the "
            "Database",
    "72": "Invalid/Missing Subscriber/Insured ID",
    "73": "Invalid/Missing Subscriber/Insured Name",
    "74": "Invalid/Missing Subscriber/Insured Gender Code",
    "75": "Subscriber/Insured Not Found",
    "76": "Duplicate Subscriber/Insured ID Number",
    "77": "Subscriber Found, Patient Not Found",
    "78": "Subscriber/Insured Not in Group/Plan Identified",
    "79": "Invalid Participant Identification",
    "80": "No Response received - Transaction Terminated",
    "81": "Invalid or Missing Case Number",
    "82": "Not Medically Necessary",
    "83": "Level of Care Not Appropriate",
    "84": "Certification Not Required for this Service",
    "85": "Certification Responsibility of External Review Organization",
    "86": "Primary Care Service",
    "87": "Exceeds Plan Maximums",
    "88": "Non-covered Service",
    "89": "No Prior Approval",
    "90": "Requested Information Not Received",
    "91": "Duplicate Request",
    "92": "Service Inconsistent with Diagnosis",
    "95": "Patient Not Eligible",
    "96": "Pre-existing Condition",
    "97": "Invalid or Missing Provider Address",
    "98": "Experimental Service or Procedure",
    "AB": "Air Brakes - Inoperative, etc.",
    "AD": "Accident Damage - Derail/Sideswiped",
    "AR": "Arrivals",
    "BA": "Reject Due to Air Bags",
    "BC": "Reject Due to No Chains",
    "BD": "Reject Due to Damps",
    "BG": "Reject Due to Bearings",
    "BL": "Reject Due to Load Divider Bad Order",
    "BP": "Reject Due to Bridge Plate",
    "BR": "Reject Due to Brake Rigging Beam, Lever",
    "BS": "Reject Due to Bad Order Slides",
    "BV": "Reject Due to Bad Order Valves/Piping",
    "BW": "Reject Due to Bad Order Walls",
    "CA": "Reject Due to Crank Arm Application",
    "CB": "Reject Due to Center Bowls, Plates and Pins",
    "CL": "Complete Loading",
    "CM": "Released as Railroad Company Material",
    "CP": "Put on Constructive Placement",
    "CS": "Release Load Through Bill Connecting Road",
    "CT": "Release as a Cross-town Load",
    "CU": "Equipment Not Used",
    "DD": "Reject Due to Doors",
    "DG": "Reject Due to Draft Gear - Yoke",
    "DM": "Dismantle",
    "DP": "Departed",
    "DR": "Reject Due to Dirty",
    "DV": "Reject Due to Load Dividers, Side Filters, Special Equipment",
    "E1": "Requested Record Will Not Be Sent; Cannot Identify the Record",
    "E2": "Requested Record Will Not Be Sent; Need Student or Parent "
            "Permission",
    "E3": "Requested Record Will Not Be Sent",
    "E4": "Requested Record Will Not Be Sent; Never Enrolled",
    "E5": "Requested Record Will Not Be Sent; No Degree Awarded",
    "E6": "Requested Record Will Not Be Sent; No Grades Posted",
    "E7": "Requested Record Cannot Be Sent Electronically; Record Resides in "
            "Paper Format only which Will Be Sent by Mail",
    "E8": "Requires Medical Review",
    "EA": "Empty Equipment Available for Loading",
    "ER": "Reject Due to Spotted in Error",
    "ET": "Empty Trailer Flat Release",
    "FR": "Reject Due to Bad Floor",
    "GS": "Release From Demurrage and Start Storage until Waybilled",
    "HB": "Reject Due to Handbrake",
    "HH": "Reject Due to Hand Hold, Ladder, Step, Running Boards, Platforms, "
            "etc.",
    "HX": "Reject Due to Hot Journal Box",
    "ID": "Releases an Idler",
    "II": "Industrial Interchange",
    "IN": "Inter-plant Switch",
    "KR": "Reject Due to Couplers",
    "LK": "Reject Due to Leaking Contents",
    "LS": "Reject Due to Load Shifted",
    "LW": "Light Weigh and Restencil",
    "MO": "Move from Current Spot to Next",
    "OG": "Reject Due to Outlet Gate/Valve Lo and Open Hopper",
    "OI": "Released from Industry to be Inspected",
    "OR": "Ordered for Replacement",
    "OV": "Reject Due to Overloaded",
    "PF": "Released Repaired by Private Facility",
    "PM": "Preventitive Maintenance",
    "PN": "Released not Repaired by Private Facility",
    "RB": "Released from Industry to Custody of Broker",
    "RD": "To be Reloaded",
    "RF": "Reject Due to Refrigeration Unit",
    "RH": "Reject Due to Roof Hatches",
    "RK": "Reject Due to Racks (Bi or Tri Levels)",
    "RL": "Released",
    "RN": "Bad Order Reinitialing and Numbering",
    "RP": "Released Partially Unloaded",
    "RS": "Released Loaded for Line Haul Shipment",
    "RT": "Run Through Equipment not Spotted",
    "SC": "Released from Demurrage after being Scrapped",
    "ST": "Released from Shop Track",
    "SU": "Reject Due to Superstructure - End, Roof and Sides",
    "SW": "Local Waybill",
    "T1": "Cannot Identify Provider as TPO (Third Party Organization) "
            "Participant",
    "T2": "Cannot Identify Payer as TPO (Third Party Organization) "
            "Participant",
    "T3": "Cannot Identify Insured as TPO (Third Party Organization) "
            "Participant",
    "T4": "Payer Name or Identifier Missing",
    "T5": "Certification Information Missing",
    "T6": "Claim does not contain enough information for re-pricing",
    "TC": "Bad Order to Transfer Lading",
    "TD": "Reject Due to Tie Down Devices",
    "TH": "Reject Due to Trailer Hitch",
    "TL": "Reject Due to Train Line, Air Hose, Anglecock",
    "TR": "Reject Due to Truck, S-Frame, Bolster",
    "UC": "Reject Due to Uncoupling Rod",
    "UF": "Reject Due to Underframe - Including Sills",
    "UG": "Bad Order for Upgrading of Car",
    "WA": "Reject Due to Wheel/Axle",
    "WK": "Bad Order Due to Wreck",
    "ZZ": "Mutually Defined",
}

eligibility_or_benefit_code = {
    "1": "Active Coverage",
    "2": "Active - Full Risk Capitation",
    "3": "Active - Services Capitated",
    "4": "Active - Services Capitated to Primary Care Physician",
    "5": "Active - Pending Investigation",
    "6": "Inactive",
    "7": "Inactive - Pending Eligibility Update",
    "8": "Inactive - Pending Investigation",
    "A": "Co-Insurance",
    "B": "Co-Payment",
    "C": "Deductible",
    "CB": "Coverage Basis",
    "D": "Benefit Description",
    "E": "Exclusions",
    "F": "Limitations",
    "G": "Out of Pocket (Stop Loss)",
    "H": "Unlimited",
    "I": "Non-Covered",
    "J": "Cost Containment",
    "K": "Reserve",
    "L": "Primary Care Provider",
    "M": "Pre-existing Condition",
    "MC": "Managed Care Coordinator",
    "N": "Services Restricted to Following Provider",
    "O": "Not Deemed a Medical Necessity",
    "P": "Benefit Disclaimer",
    "Q": "Second Surgical Opinion Required",
    "R": "Other or Additional Payor",
    "S": "Prior Year(s) History",
    "T": "Card(s) Reported Lost/Stolen",
    "U": "Contact Following Entity for Eligibility or Benefit Information",
    "V": "Cannot Process",
    "W": "Other Source of Data",
    "X": "Health Care Facility",
    "Y": "Spend Down"
}



from tigershark import X12_4010_X061A1
from tigershark import X12_4010_X070
from tigershark import X12_4010_X091A1
from tigershark import X12_4010_X092A1
from tigershark import X12_4010_X093A1
from tigershark import X12_4010_X094A1
from tigershark import X12_4010_X095A1
from tigershark import X12_4010_X096A1
from tigershark import X12_4010_X097A1
from tigershark import X12_4010_X098A1
from tigershark import X12_4010_XXXC
from tigershark import X12_5010_X221A1
from tigershark import X12_5010_X279A1
from tigershark.extras import standardSegment
from tigershark.X12.parse import Loop
from tigershark.X12.parse import Message
from tigershark.X12.parse import ParseError
from tigershark.X12.parse import Properties

# Transaction Set ID -> X12VersionTuple -> [(parser module, parser name)]
PARSER_MAP = {
    '270': {
        X12_4010_X092A1: [('M270_4010_X092_A1', 'parsed_270')],
        X12_5010_X279A1: [('M270_5010_X279_A1', 'parsed_270')],
    },
    '271': {
        X12_4010_X092A1: [('M271_4010_X092_A1', 'parsed_271')],
        X12_5010_X279A1: [('M271_5010_X279_A1', 'parsed_271')],
    },
    '276': {
        X12_4010_X093A1: [('M276_4010_X093_A1', 'parsed_276')],
    },
    '277U': {
        X12_4010_X070: [('M277U_4010_X070', 'parsed_277U')],
    },
    '277': {
        X12_4010_X093A1: [('M277_4010_X093_A1', 'parsed_277')],
    },
    '278': {
        X12_4010_X094A1: [('M278_4010_X094_27_A1', 'parsed_278'),
                          ('M278_4010_X094_A1', 'parsed_278')],
    },
    '820': {
        X12_4010_X061A1: [('M820_4010_X061_A1', 'parsed_820')],
    },
    '834': {
        X12_4010_X095A1: [('M834_4010_X095_A1', 'parsed_834')],
    },
    '835': {
        X12_4010_X091A1: [('M835_4010_X091_A1', 'parsed_835')],
        X12_5010_X221A1: [('M835_5010_X221_A1', 'parsed_835')],
    },
    '837': {
        X12_4010_X096A1: [('M837_4010_X096_A1', 'parsed_837')],
        X12_4010_X097A1: [('M837_4010_X097_A1', 'parsed_837')],
        X12_4010_X098A1: [('M837_4010_X098_A1', 'parsed_837')],
    },
    '841': {
        X12_4010_XXXC: [('M841_4010_XXXC', 'parsed_841')],
    },
}


def get_parsers(transaction_set_id, version_tuple):
    """
    Generate the parsers to try for a given transaction_set and version.

    The parsers should be tried in the given order.
    """
    try:
        parsers = PARSER_MAP[transaction_set_id][version_tuple]
    except KeyError:
        raise ValueError("Unsupported transaction set and version.",
                         transaction_set_id, version_tuple)

    for index, (module_name, parser_name) in enumerate(parsers):
        module = __import__('tigershark.parsers.' + module_name,
                            fromlist=[parser_name])
        yield getattr(module, parser_name)


class SimpleParser(object):
    """
    A parser for a particular transaction set and version.
    """
    def __init__(self, transaction_set_id, version_tuple):
        self.transaction_set_id = transaction_set_id
        self.version_tuple = version_tuple

    def unmarshall(self, x12_contents, **kwargs):
        for parser in get_parsers(self.transaction_set_id, self.version_tuple):
            try:
                return parser.unmarshall(x12_contents, **kwargs)
            except ParseError:
                continue
        else:
            # Let the last parser raise.
            return parser.unmarshall(x12_contents, **kwargs)


STLoop = Loop(
    u'ST_LOOP',
    Properties(
        position=u'0200',
        looptype=u'explicit',
        repeat=u'>1',
        req_sit=u'R',
        desc=u'Transaction Set Header',
    ),
    standardSegment.st,
)

GSLoop = Loop(
    u'GS_LOOP',
    Properties(
        position=u'0200',
        looptype=u'explicit',
        repeat=u'>1',
        req_sit=u'R',
        desc=u'Functional Group Header',
    ),
    standardSegment.gs,
    STLoop,
)

ISALoop = Loop(
    u'ISA_LOOP',
    Properties(
        position=u'0010',
        looptype=u'explicit',
        repeat=u'>1',
        req_sit=u'R',
        desc=u'Interchange Control Header',
    ),
    standardSegment.isa,
    GSLoop,
)

ControlParser = Message(
    u'ASC X12 Interchange Control Structure',
    Properties(
        desc=u'ASC X12 Control Structure',
    ),
    ISALoop,
)


def parse_control_headers(x12_contents):
    """Parse only the control headers of an X12 message.

    This parses the three outer control headers of the given X12 message:

    * Interchange Control (ISA)
    * Functional Group (GS)
    * Transaction Set (ST).

    These can be used to identify the X12 versions and transaction set types
    contained in the message.
    """
    return ControlParser.unmarshall(x12_contents, ignoreExtra=True)



from tigershark.facade import get_facade
from tigershark.facade.control import ControlHeaders
from tigershark.parsers import SimpleParser
from tigershark.parsers import parse_control_headers


def identify_simple_x12(x12_contents):
    """
    Return (transaction set id, version tuple) for simple X12.

    Simple X12 has only one kind of transaction set and one version.

    :param x12_contents: A string with X12 content.
    :returns: (transaction set id, version_tuple)
    :raises ValueError: If we cannot find any transaction set.
    :raises ValueError: If we find more than one kind of transaction set
        or version.
    :raises ParseError: If the X12 cannot be parsed.
    """
    identifers = set()  # (transaction set id, version tuple)

    headers = ControlHeaders(parse_control_headers(x12_contents))

    for control in headers.interchange_controls:
        for group in control.functional_groups:
            for transaction_set in group.transaction_sets:
                identifers.add((
                    transaction_set.transaction_set_identifier_code,
                    group.version_tuple))

    if not identifers:
        raise ValueError('No transaction sets')
    elif len(identifers) > 1:
        raise ValueError('Multiple kinds of transaction sets.', identifers)
    else:
        return identifers.pop()


def apply_facade_to_simple_x12(x12_contents):
    """
    Identify, parse, and apply a facade to simple X12.

    If the X12 can be identified and parsed, the appropriate
    Facade is applied and returned.

    :param x12_contents: A string with X12 content.
    :returns: A Facade instance.
    :raises ValueError: If we cannot find any transaction set.
    :raises ValueError: If we find more than one kind of transaction set
        or version.
    :raises ValueError: If we cannot find an appropriate Facade.
    :raises ParseError: If the X12 cannot be parsed.
    """
    transaction_set_id, version_tuple = identify_simple_x12(x12_contents)
    parser = SimpleParser(transaction_set_id, version_tuple)
    x12 = parser.unmarshall(x12_contents)
    return get_facade(transaction_set_id, version_tuple)(x12)



#!/usr/bin/env python
"""Top-level application programs.

-   :mod:`tools.convertPyX12`.
    Converts Python :file:`.xml` files to :mod:`X12.parse` structures.
    From these structures visitors in :mod:`X12.map` can emit Python source
    as well as Django ORM structures.

-   :mod:`tools.surveyPyX12`.
    Examines the Python :file:`.xml` files to write some overview
    reports.

tools.convertPyX12
=====================

..  automodule:: tools.convertPyX12
    :members:

tools.surveyPyX12
===================

..  automodule:: tools.surveyPyX12
    :members:


"""
